#include "COPY"

/* -------------------------------------------------------------------------
   | Funktion zum ersetzen von Umlauten in Text-                           |
   | dateien durch "GML-Zeichen"                                           |
   |                                                                       |
   | Autor: Thomas Wendl (rz6b@dv.op.dlr.de)                               |
   | Version: 1.13 vom 25.02.1993                                          |
   | 2-sprachige Meldungen: Kre, 8.4.1993                                  |
   |-----------------------------------------------------------------------|
   | Aufruf: char_convert( tabelle, quelle, ziel )                         |
   |                 char *tabelle               Zeiger auf den Namen einer|
   |                                             Datei, die eine Ueber-    |
   |                                             setzungs-Tabelle enthaelt |
   |                                             ( NULL fuer default-Tabel-|
   |                                                                  le ) |
   |                          char *quelle       Zeiger auf Namen der      |
   |                                             Quelldatei ( "-w" ==      |
   |                                                  Ueberschreib-Modus ) |
   |                                  char *ziel Zeiger auf Namen der      |
   |                                                            Ziel-Datei |
   |                                                                       |
   |                         ( quelle == ziel ---> Ueberschreib-Modus )    |
   ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include "char_conv.h"          /* Meldungen fr char_conv, deutsch und engl */

typedef int BOOL;

BOOL inizialize(char arg[]);
                       /* Funktion zum Bestimmen der UEbersetzung-Tabelle */

void closeall(void);   /* Funktion zum Schliessen aller angeforderten Dateien*/

int cnvrtgml(char input[], volatile char **output);
             /* Funktion zur Umwandlung einer Zeile mit Umlauten in eine "GML-
                Zeile". UEbergabe der Original-Zeile und eines ausreichenden
                Puffers fuer die Konvertierung.   Rueckgabe eines "Erfolgs-
                Werts" (BOOL bedeutet Bool'sche oder logische Variable) */

BOOL I_want_more(char **buffer); /* Funktion zum Vergroessern des Speichers der
                                   GML-Zeile */

void corrupt(char arg[],char msg[]); /* Fehler IN der UEbersetzungs-Tabelle */

void no_mem(void);                  /* Beschwerde bei Speicher-Mangel */

static int lg;                     /* Sprache der Meldungen */

FILE *uxdatei;                    /* Eingabe-Datei */
FILE *gmldatei;                  /* Ausgabe-Datei */
FILE *table;                    /* Tabellen-Datei */
unsigned buffer_size=256;      /* Groesse der Ausgabezeile (default) */
char **replacement, *to_find; /* Suchen und Ersetzen */
char *uxzeile,*gmlzeile;     /* Eingabe und Ausgabe */
char conv_buffer[2048];     /* Puffer fuer zurueckgegeben Meldungen */ 
int number,satzlaenge;     /* Anzahl der 'Zeichen' und der pro Zeile */

int cnvrtgml(char input[], volatile char **output)
 /* Uebergabe der Zeile mit Umlaute, Rueckgabe als Zeile mit "GML-Umlauten" */
  {
  int success=-1;
  BOOL replaced=FALSE;
  int i=0,j=0,k,l;
  while (input[i]!='\0')
    {
    l=0;
    for (k=0;k<number;k++)
      {
      if (input[i]==to_find[k])
        {
        while (replacement[k][l]!='\0') 
          output[0][j++]=replacement[k][l++];
        replaced=TRUE;
        k=number;
        }
      }
    if (replaced) replaced=FALSE;
    else output[0][j++]=input[i];
    i++;
    if (j>=buffer_size) 
    {  if (!I_want_more(output)) return FALSE; }  /* Zu wenig Speicher?? */
    };                                           /* Dann hol' noch welchen!! */
  if (j>=satzlaenge) success=j;
  output[0][j]=input[i];
  return success;                /* Zeile abgearbeitet -> Erfolg mitteilen */
  };



void closeall(void)
  {
  if (table!=NULL)
   { fclose(table); table=NULL; }
  if (uxdatei!=NULL)
   { fclose(uxdatei); uxdatei=NULL; }
  if (gmldatei!=NULL)
   { fclose(gmldatei); gmldatei=NULL; }
  if (uxzeile!=NULL)
   { free(uxzeile); uxzeile=NULL; }
  if (gmlzeile!=NULL)
   { free(gmlzeile); gmlzeile=NULL; }
  if (to_find!=NULL)
   { free(to_find); to_find=NULL; }
  if (replacement!=NULL)
    {
    int a;
    for (a=0;a<number;a++)
    { if (replacement[a]!=NULL) free(replacement[a]); }
    free (replacement); replacement=NULL;
    }
  };



BOOL I_want_more(char **buffer)
  {
  buffer_size+=128;
  buffer[0]=realloc(buffer[0],buffer_size+11);
  if (buffer[0]!=NULL) return TRUE;
  no_mem();
  return FALSE;
  };




void no_mem(void)
  {
  strcat(conv_buffer,&NOMEMFREE[lg][0]);
  closeall();
  };
 


BOOL inizialize(char arg[])
  {
  int length,index;
  char raw[256];
  int a,i;
  char de_faults[18][15]={" &a."," &A."," &o."," &O.",
                       " &u."," &U."," &s.",
                       "! &exclm.","~ &tilde.","| &vbar.",
                       "{ &lbrace.","} &rbrace."," &acute.",
                       "[ &lbracket.","] &rbracket.","` &grave.",
                       "^ &circumflex.","\\ &bsl."};
  
                       /* arg != \0 ===> Tabelle laden */ 

  if (0!=strcmp(arg,""))
    {
    table=fopen(arg,"r");
    if (table==NULL) 
      { strcat(conv_buffer,&TABLE[lg][0]);
        strcat(conv_buffer,arg);
        strcat(conv_buffer,&TABLE2[lg][0]);
        return FALSE; }
    number=-1;
    while (!feof(table))
      {
      if (fgets(raw,256,table)==NULL) break;
      if (((raw[0]!='\n')&&(raw[0]!='#'))||((raw[0]=='#')&&(raw[1]=='#')))
        number++;
      } 
    rewind(table);
    to_find=malloc(number);
    replacement=calloc(number,sizeof(char*));
    if ((!to_find)||(!replacement)) { no_mem(); return FALSE; }

           /* ^ Erste Lesung + Speicher bestellen ^ */
           /* Jetzt kommt: Zweite Lesung == Werte einlesen */

    fgets(raw,256,table);
    if (1!=sscanf(raw,"maxlength=%d",&satzlaenge))
      { corrupt(arg,&CORRUPT_MSG_1[lg][0]); return FALSE; }
    for(a=0;a<number;a++)
      {
      length=0;
      do { fgets(raw,256,table); 
      } while ((raw[0]=='\n')||((raw[0]=='#')&&(raw[1]!='#')));
      for (i=0;isspace(raw[i])!=0;i++) ;
      index=i+1;
      if (raw[i]=='#')                          { to_find[a]='#'; index++; }
      else if (raw[i]!='\\')                      to_find[a]=raw[i];
      else if ((raw[i+1]=='\\')&&(raw[i+2]==' '))   { to_find[a]='\\'; index++;}
      else if ((isdigit(raw[i+1])!=0)&&(isdigit(raw[i+2])!=0)&&(isdigit(raw[i+3])!=0)&&(raw[i+4]==' '))
       {
              /*  Hoppla!? Oktal-Schreibweise?!? */

       to_find[a]=0;
         for (index=i+1;index<4+i;index++)     /* "Horner" laesst gruessen */
         {  int c=to_find[a]; 
            raw[index]=raw[index]-48;
            c*=8;
            if ((0<=raw[index])&&(raw[index]<8))  c+=raw[index];
            else { corrupt(arg,&CORRUPT_MSG_2[lg][0]); return FALSE; }
            if (c>255)
              { corrupt(arg,&CORRUPT_MSG_2[lg][0]); return FALSE; }
            to_find[a]=c; }
       }

      else if ((isalnum(raw[i+1])!=0)&&(raw[i+2]==' '))  
        {
              /* Ahhh!! Ein 'White-Space' in C-Notierung! */

        if (raw[i+1]=='t')                        { to_find[a]='\t'; index++;}
        else if (raw[i+1]=='n')                   { to_find[a]='\n'; index++;}
        else { corrupt(arg,&CORRUPT_MSG_3[lg][0]); return FALSE; }       /* \0 ist nicht moeglich */
        }
      else { corrupt(arg,&CORRUPT_MSG_3[lg][0]); return FALSE; }
      while (0!=isspace(raw[index]))          /* Blanks, etc ausmaskieren*/
        { if ((raw[index]=='\n')||(raw[index++]=='\0')) 
            { corrupt(arg,&CORRUPT_MSG_0[lg][0]); return FALSE; }  }
      while (0==isspace(raw[index+length]))
        {
        int b=index+length;
        if (raw[b]!='\\')  ;
        else if (raw[b+1]=='\\') strcpy(raw+b+1,raw+b+2);
        else if ((isdigit(raw[b+1])!=0)&&(isdigit(raw[b+2])!=0)&&(isdigit(raw[b+3])!=0))
          {
                        /* Und nochmal: Oktal a la Horner */
          int d;
          raw[b]=0;
          for (d=1;d<4;d++)
           { int c=raw[b]; 
             raw[b+d]=raw[b+d]-48;
             c*=8;
             if ((0<=raw[b+d])&&(raw[b+d]<8))  c+=raw[b+d];
             else { corrupt(arg,&CORRUPT_MSG_2[lg][0]); return FALSE; }
             if (c>255) { corrupt(arg,&CORRUPT_MSG_2[lg][0]); return FALSE; }
             raw[b]=c; }
          strcpy(raw+b+1,raw+b+4);
          }
        else
          {
                        /* Nochmal ein 'White-Space' \n\n */
          if (raw[b+1]=='t')                         raw[b]='\t';
          else if (raw[b+1]=='n')                    raw[b]='\n';
          else { corrupt(arg,&CORRUPT_MSG_3[lg][0]); return FALSE; }
          strcpy(raw+b+1,raw+b+2);              /* \0 ist im replacement verboten!!*/
          } 
        length++;
        }
      replacement[a]=malloc(length);
      if (replacement[a]==NULL) { no_mem(); return FALSE; }
      strncpy(replacement[a],raw+index,length);
      replacement[a][length]='\0';
      }
    fclose(table);
    table=NULL;
    }
  else                 /* arg == \0 ===> default-table einlesen */
    {
    number=18;
    length=12;
    satzlaenge=256;
    to_find=malloc(number);
    replacement=calloc(number,sizeof(char*));
    if ((!to_find)||(!replacement)) { no_mem(); return FALSE; }
    for(a=0;a<number;a++)
     {  replacement[a]=malloc(length); 
        if (replacement[a]==NULL) { no_mem(); return FALSE; }
        sscanf(de_faults[a],"%c %s",&(to_find[a]),replacement[a]); }
    }
  return TRUE;
  };



void corrupt(char arg[],char msg[])
  {
  strcat(conv_buffer,msg);
  strcat(conv_buffer,&CORRUPT_TABLE[lg][0]);
  strcat(conv_buffer,arg);
  strcat(conv_buffer,&CORRUPT_TABLE2[lg][0]);
  closeall();
  };



 /* und nun kommt char_convert */

char *char_convert(char *tabelle,char *quell,char *ziel,int language)
  {
  char tmp_file[100], *Pfadende;    /* Zur Erzeugung einer Tmp-Datei */
  char satz_char[15],zeile_char[15];
  BOOL ueberlaenge=FALSE;
  struct stat FileDir1[1];
  struct stat FileDir2[1];
  int i=1,j;
  int FileTest;
  int argi;                          /* Argument-Index */
  int FileFlags;
  int conv_success;

  lg = language;        /* wir nehmen die globale Variable, damit wir die
			   Sprache nicht berall hin mit bergeben mssen */
  conv_buffer[0]='\0';
  if (!inizialize(tabelle)) return conv_buffer;
  if ((0==strcmp(quell,"-w"))||(0==strcmp(quell,ziel))) 
    { quell=ziel;
      Pfadende=strrchr(quell,'/');
      if (Pfadende>quell)  strncpy(tmp_file,quell,(Pfadende-quell)+1);
      strcat(tmp_file,"gmlconvtmp");  }

             /* Open Open Open Open Open Open */

  stat(quell,FileDir1);           /* Status in FileDir1 speichern */
  if (((FileDir1->st_mode)&S_IFMT)!=S_IFREG) uxdatei=NULL;
  else uxdatei=fopen(quell,"r");  /*Stream nur oeffnen wenn Wirklich FILE*/
  if (uxdatei==NULL)                /*S_IFREG == File, S_IFMT == Typen-Maske*/
    {
    strcat(conv_buffer,&SOURCE[lg][0]);
    strcat(conv_buffer,quell);
    strcat(conv_buffer,&SOURCE2[lg][0]);
    return conv_buffer;
    }
  if (0==strcmp(ziel,quell)) gmldatei=fopen(tmp_file,"w");
  else
    {
    if ((stat(ziel,FileDir2)==0)&&(((FileDir2->st_mode)&S_IFMT)!=S_IFREG))
       gmldatei=NULL;
    else  gmldatei=fopen(ziel,"w");
    }
/* Erklaerung: Stream nur oeffnen wenn 1. FILE
                                  oder 2. Status laden scheiterte.
                                      weil a) File nicht existiert
                                             ---> fopen erzeugt neues File
                                      oder b) Keine Schreibberechtigung
                                             ---> fopen gibt NULL zurueck */
  if (gmldatei==NULL)
    {
    strcat(conv_buffer,&DEST[lg][0]);
    if (quell==ziel)  strcat(conv_buffer,tmp_file);
    else strcat(conv_buffer,ziel);
    strcat(conv_buffer,&DEST2[lg][0]);
    closeall();
    return conv_buffer;
    }

       /* Erstbelegung des Speichers */

  gmlzeile=malloc(buffer_size+11);
  if (satzlaenge==0)
    uxzeile=malloc(257);
  else 
    uxzeile=malloc(satzlaenge+1);
  if ((gmlzeile==NULL)||(uxzeile==NULL))  
   { no_mem(); return conv_buffer; }

        /* Lesen, Verarbeiten, Schreiben */

  while (feof(uxdatei)==0)
    {
    j=0;
    if (satzlaenge==0)
      fgets(uxzeile,257,uxdatei);      /* Lesen der Zeile */
    else
      {
      fgets(uxzeile,satzlaenge+1,uxdatei);
      if (uxzeile[strlen(uxzeile)-1]!='\n')
        { strcat(conv_buffer,&LONGLINE_READ[lg][0]);
          sprintf(satz_char,"%d",satzlaenge);
          strcat(conv_buffer,satz_char);
          strcat(conv_buffer,&LONGLINE_READ2[lg][0]);
          closeall();
          return conv_buffer;  }
      }
    if (feof(uxdatei)!=0) break;
    conv_success=cnvrtgml(uxzeile, &gmlzeile);
    if (!conv_success) return conv_buffer;
    else if ((satzlaenge)&&(conv_success>satzlaenge))
      {
      strcat(conv_buffer,&LONGLINE_WRITE[lg][0]);
      sprintf(zeile_char,"%d",i);
      strcat(conv_buffer,zeile_char);
      strcat(conv_buffer,&LONGLINE_WRITE2[lg][0]);
      ueberlaenge=TRUE;
      }
    fprintf (gmldatei,"%s",gmlzeile);
    i++;
    }
  closeall();

      /* Ende ENTE Fin End ..... */

  if (ueberlaenge)
    strcat(conv_buffer,&LONGLINE_WRITE_END[lg][0]);
  if (0==strcmp(ziel,quell))
    {
    if ((remove(quell)!=0)||(rename(tmp_file,quell)!=0))
      {
      strcat(conv_buffer,&REPLACE[lg][0]);
      strcat(conv_buffer,tmp_file);
      strcat(conv_buffer,&REPLACE2[lg][0]);
      return conv_buffer;
      }
    }
  if (conv_buffer[0]=='\0') return &SUCCESS[lg][0];
  else return conv_buffer;
  };
