#include "COPY"

/*
 * MOTeX 0.9 (Motif Oberflche fr TeX), 1993
 * 
 * author: Klaus Kretschel, DLR (rz13@vm.op.dlr.de)
 *
 * based on: WCL 2.4
 *           AIX TeX as from ftp@rusinfo.rus.uni-stuttgart.de
 *
 * change log:
 *
 * 23.04.93 Kre      German/Englisch via resource or option xnlLanguage
 * 06.05.93 Kre      support dvilj/2/p family
 */

#include <X11/Xm/Xm.h>		/* Motif and Xt Intrinsics	*/
#include <X11/StringDefs.h>	/* Definitionen fr resRecord	*/
#include <X11/Wc/WcCreate.h>	/* Widget Creation Library	*/
#include <X11/Xmp/Xmp.h>	/* Motif Public widgets etc.	*/
#include <signal.h>		/* 	*/
#include <locale.h>		/* 	*/
#include "motex.h"	        /* Header File fr TeX Shell */
#include "resources.h"	        /* resource Definition fr TeX Shell */

#define WcNull(str)  (int)(!((str) && *(str)))  /* aus WcCreate.h */

#define  PP_BUFLEN       64     /* maximum number of private printers */

/******************************************************************************
**  Private Data
******************************************************************************/

int msg_lg;

/* All Wcl applications should provide at least these Wcl options:
*/
static XrmOptionDescRec options[] = {
  WCL_XRM_OPTIONS,
  TEXSHELL_OPTIONS
  };

XtAppContext  app;	 /* auch in den CBs gebraucht */
char	      *home,	 /* Home-Directory */
	      *pwd,	 /* current working directory */
	      *texdir;   /* directory tex in home (if present) */

AppData  app_data;
char    *appName;

/* die enthalten den jeweilig gerade gltigen default Editor, ...:
 */
Widget  default_edit, default_format, default_view, default_drive;

static XtResource tx_rc[] = {
  {
    XmNxnlLanguage, XmCXnlLanguage, XmRString, sizeof(String),
    XtOffsetOf(AppData, lg), XmRString, NULL
  },
  {
    XmNxtermCommand, XmCXtermCommand, XmRString, sizeof(String),
    XtOffsetOf(AppData, xterm), XmRString, XTERM_CMD
  },
  {
    XmNbackground, XmCBackground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData, bg), XtRString, XtDefaultBackground
  },
  {
    XmNforeground, XmCForeground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData, fg), XtRString, XtDefaultForeground
  },
  {
    XmNattentionColor, XmCBackground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData, hi_col), XtRString, XtDefaultBackground
  },
  {
    XmNfsbBackground, XmCBackground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData, fsb_bg), XtRString, XtDefaultBackground
  },
  {
    XmNdefaultEditor, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_edit), XtRString, "emacs"
  },
  {
    XmNdefaultFormatter, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_form), XtRString, "LaTeX"
  },
  {
    XmNdefaultPreviewer, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_view), XtRString, "xdvi"
  },
  {
    XmNdefaultDriver, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_drive), XtRString, "dvips"
  },
  {
    XmNdefaultPrinter, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_printer), XtRString, NULL
  },
  {
    XmNprivatePrinterList, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, add_printer), XtRString, NULL
  },
  {
    XmNprimaryInputFileName, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_infile1), XtRString, NULL
  },
  {
    XmNalternateInputFileName, XmCMenuItem, XmRString, sizeof(String),
    XtOffsetOf(AppData, df_infile2), XtRString, NULL
  }
};

/* Liste der Widgets, die im customize Men gesteuert werden knnen    */
/* optionWidgetName: Name des Widgets im Option Men, gleichzeitig     */
/*                   resource value fr den Benutzer (gro oder klein),*/
/*                   wenn man den * weg lt.                          */
/* optionWidget:     Widget-Ids dazu, werden in setDefaultToggleButtons*/
/*                   besetzt                                           */
/* custWidgetName:   Name des Widgets im customize Men                */

_list    editors[] = { {"*emacs",    NULL, "*e3"},
		       {"*INed",     NULL, "*e2"},
		       {"*vi",       NULL, "*e1"} },
      formatters[] = { {"*LaTeX",    NULL, "*f1"},
		       {"*SLiTeX",   NULL, "*f2"},
		       {"*Plain",    NULL, "*f3"} },
      previewers[] = { {"*xdvi",     NULL, "*v1"},
		       {"*ghostview",NULL, "*v2"} },
         drivers[] = { {"*dvips",    NULL, "*p1"},
		       {"*dvilj",    NULL, "*p2"} };

/* weil XtNumber nur da benutzt werden kann, wo die Liste deklariert wird: */

int editor_count    = XtNumber(editors),
    formatter_count = XtNumber(formatters),
    previewer_count = XtNumber(previewers),
    driver_count    = XtNumber(drivers); 

char   *add_printers[PP_BUFLEN],  /* list of private printer names */
       *add_prt_list;             /* same in string format */
int     add_printer_count;

char   *English,                  /* Locale string for English and */
       *German;                   /* German; to be set in check_language */

extern void	execute();
extern void	setDir();
extern void	setFileName();
extern void     setTextString();
extern void	setOption();
extern void	setWidget();
extern void	setDefaultToggleButton();
extern void	setDefaultAccelerators();
extern void     addPrinter();
extern void	saveValues();
extern void	nextField();
extern void	showHelp();
extern void	reset();
extern Boolean	a_eq_b(char *, char *);
extern char    *strip(char *, char);

/*ARGSUSED*/
static void RegisterApplication ( app )
{
    /* -- Useful shorthand for registering things with the Wcl library */
#define RCP( name, class  ) WcRegisterClassPtr   ( app, name, class );
#define RCO( name, constr ) WcRegisterConstructor( app, name, constr );
#define RAC( name, func   ) WcRegisterAction     ( app, name, func );
#define RCB( name, func   ) WcRegisterCallback   ( app, name, func, NULL );

    /* -- register widget classes and constructors */
    /* -- Register application specific actions */
    /* -- Register application specific callbacks */
    RCB ( "execute", execute )
    RCB ( "setFileName", setFileName )
    RCB ( "setDir", setDir )
    RCB ( "setOption", setOption )
    RCB ( "setWidget", setWidget )
    RCB ( "setDefaultToggleButton", setDefaultToggleButton)
    RCB ( "saveValues", saveValues)
    RCB ( "setDefaultAccelerators",setDefaultAccelerators)
    RCB ( "addPrinter",addPrinter)
    RCB ( "setTextString",setTextString)
    RCB ( "nextField", nextField)
    RCB ( "showHelp", showHelp)
}

/*----------------------------------------------------------------*
 * berprft, ob die angegebene resource gltig ist und setzt die
 * zugehrigen Widget Ids; gibt Index in die Liste zurck. 
 *----------------------------------------------------------------*/
int checkDefaults( Widget w, char * in, _list list[], int count )
{
  int i;
  for (i=0;i<count;i++)
    list[i].optionWidget = WcFullNameToWidget(w, list[i].optionWidgetName);
  for (i=0;i<count;i++)
    if (a_eq_b(in,list[i].optionWidgetName+1)) return i;

  printf("invalid resource value \"%s\", \"%s\" used\n",
	 in,list[0].optionWidgetName+1);
  return 0;
}

/*-------------------------------------------------------------*
 * sets app_data.df_.. defaults from resource files;
 * must not be called before all needed widgets are known
 *-------------------------------------------------------------*/
#define CHK(a,b,c)  i = checkDefaults(w, app_data.a,c,XtNumber(c)); \
                    app_data.a = c[i].custWidgetName; \
                    b = c[i].optionWidget;

void checkResources( Widget w )
{
  int i;
  CHK(df_edit, default_edit,  editors)
  CHK(df_form, default_format,formatters)
  CHK(df_view, default_view,  previewers)
  CHK(df_drive,default_drive, drivers)
}     

/*--------------------------------------------------------------------*
 * resource privatePrinterNames checken und aufbereiten; ein Name
 *  ist ok, wenn ein config file dazu in DEFAULT_PATH gefunden wird.
 *--------------------------------------------------------------------*/

int checkPrinterList(char *list)
{
  int   c = 0;
  char path[1024], *filename;
  
  add_printers[c] = strtok(list," ,\t");
  while ( c < PP_BUFLEN && add_printers[c] != NULL)
    {
      sprintf(path,"%s/config.%s:%s/dvips.config/config.%s",
	      home,add_printers[c],home,add_printers[c]);
      filename = XtFindFile(path,0,0,0);
      if (filename)
	add_prt_list = concat(add_prt_list,add_printers[c++]," ",NULL);
      else fprintf(stderr,"invalid private printer name \"%s\", ignored.\n",
		   add_printers[c]);
      XtFree(filename);
      add_printers[c] = strtok(NULL," ,\t");
    }
  return c;
}

int check_language(char *lg)
{

  if (WcNull(lg)) lg = getenv("LANG");      /* xnlLanguage nicht angegeben;
                                             * aendert nicht den Aufrufwert! */
  English = islower(lg[0]) ? "en_US":"En_US";  /* fr ISO/PC richten wir uns */
  German = islower(lg[0]) ? "de_DE":"De_DE";   /* nach dem bisherigen Eintrag*/

  if (a_eq_b("de_", lg)) return 1;          /* irgendein deutsch */
  else if (a_eq_b("en_", lg)) return 0;     /* irgendein englisch */
  else fprintf(stderr,"invalid resource value for xnlLanguage \"%s\".\n",lg);
  return 0;
}

/******************************************************************************
*   MAIN function
******************************************************************************/

main ( int argc, String argv[] )
{   
  Pixmap icon_pixmap;
  Screen  *screen;

  /*  -- Intialize Toolkit creating the application shell
   */
  Widget appShell = XtInitialize ( 
	argv[0], appName = WcAppClass( argc, argv ),	/* application name & class */
	options, XtNumber(options),		/* descr of cmd line options */
	&argc, argv );
  app = XtWidgetToApplicationContext(appShell);

  /* -- need home and current working directory;
   *    set name of TeX directory
   */
  home = getenv("HOME");
  pwd = getenv("PWD");
  texdir = concat(home,"/tex",NULL);

  /* -- Icon Pixmap setzen, vorlufig hard-coded
   */
  screen = XtScreen(appShell);
  icon_pixmap = XmGetPixmap(screen, MOTEX_ICON,
			    WhitePixelOfScreen(screen),
			    BlackPixelOfScreen(screen));
  XtVaSetValues(appShell, XmNiconPixmap, icon_pixmap, NULL);

  /*  -- application resources holen; in setDefaultAccelerators wird 
   *  erst gecheckt, weil die Widgets sonst noch nicht bekannt sind;
   *  printer list geht sofort 
   */

  XtGetApplicationResources(appShell,
			    &app_data, tx_rc, XtNumber(tx_rc),
			    NULL, 0);
  add_printer_count = checkPrinterList(app_data.add_printer);

  msg_lg = check_language(app_data.lg);
/* puts(XtName(appShell));
  XtVaSetValues(appShell,XmNxnlLanguage,"de_DE",NULL);  ntzt nix */

  /* input Felder sensitiv fr Umlaute setzen,
   * Kodierung wird aus LANG geholt
   */
  setlocale(LC_ALL, "");

  /*  -- Register all application specific callbacks and widget classes
   */
  RegisterApplication ( app );

  /*  -- Register all Motif classes, constructors, and Xmp CBs & ACTs
   */
  XmpRegisterAll ( app );

  /*  -- Create widget tree below toplevel shell using Xrm database
   */
  if ( WcWidgetCreation ( appShell ) )
    exit(1);

  /*  -- Realize the widget tree
   */
  XtRealizeWidget ( appShell );

  /*  -- Optional, but frequently desired:  
   * Provide a callback which gets invoked when the user selects `Close' from
   * the mwm frame menu on the top level shell.  A real application will need
   * to provide its own callback instead of DeleteWindowCB, and probably 
   * client data too.  MUST be done after shell widget is REALIZED!  Hence,
   * this CANNOT be done using wcCallback (in a creation time callback).
   */
  /* XmpAddMwmCloseCallback( appShell, DeleteWindowCB, NULL );
   */

  /* want to be notified when child programs die (editor, tex ...)
   */
  signal(SIGCHLD,reset);

  /*  -- and finally, enter the main application loop
   */
  XtMainLoop ( );
}


