#include "COPY"

/* utilities for MOTeX
 *
 * u.a. config files fr PS und HP LJ lesen und die Daten aufbereiten
 * Parameter: Liste der privaten Drucker
 */

#include <stdarg.h>
#include <ctype.h>
#include <Xm/Xm.h>
#include <Wc/WcCreate.h>
/* signal.h mu nach X, sonst wird die Definition von SIGCHLD kaputt gemacht */
#include <signal.h>
#include "motex.h"	        /* Header File fr TeX Shell, wg. _cmd */

#define BUFLEN     1024		/* Max Lnge fr diverse buffer */

extern char *START_MSG_1[],
            *START_MSG_2[],
            *INV_CONF_FILE1[],
            *INV_CONF_FILE2[];

extern int     msg_lg;
extern AppData app_data;

/* printer_list connects the printer names, as defined in the printer menu,
 * with driver name and parameters; they must be the same as in Texshell.prt;
 * the fields are: 
 *   1. widget name, same as in Texshell.prt (NOT the widget labelString)
 *   2. driver name
 *   3. driver options
 *   4. with dvilj, the output is spooled from file to the printer, usually
 *      in a pipe; what comes behind the pipe sign, is here ("qprt -P queue")
 *   5. address of next printer in list, NULL for last.
 * Currently this data is only used for Laserjet drivers.
 * There must be at least one entry.
 *
 * The address of the last available printer is held in last_printer_addr.
 */

Printer printer_list[] = {
  {"300dpi_file",   "dvips",  NULL,  NULL,  &printer_list[1] },
  {"bs6rz001",      "dvips",  NULL,  NULL,  &printer_list[2] },
  {"go6rz001",      "dvips",  NULL,  NULL,  &printer_list[3] },
  {"kp6rz001",      "dvips",  NULL,  NULL,  &printer_list[4] },
  {"op6rz001",      "dvips",  NULL,  NULL,  &printer_list[5] },
  {"op6pa004",      "dvips",  NULL,  NULL,  NULL             } };

Printer *last_printer_addr = &printer_list[5];

/* defaults werden durch wcCallback im resource-file definiert;
 *
 * each command to be invoked needs an entry in this structure;
 * fields are:
 *   1. widget id of button
 *   2. to which group button belongs:
 *         e  editor
 *         f  format package
 *         v  previewer
 *         d  printer driver
 *      others are special
 *   3. name of button, as defined in app-defaults/Texshell.main,
 *      must match XtName(widget_id)
 *   4. command name to be invoked by 3
 *   5. 1 if command is to be invoke via the -e option of xterm command,
 *      0 otherwise
 *   6. process id of child when command is invoked
 *
 * 2-5 must be supplied, 1 and 6 are set on invocation
 */

_cmd  cmd[] = {
  { NULL,   "b",        "BibTeX",       "bibtex",       1,  0 },
  { NULL,   "x",	"xfig",		"xfig",	        0,  0 },
  { NULL,   "e",	"vi",		"vi",	        1,  0 },
  { NULL,   "e",	"INed",		"e",	        0,  0 },   	
  { NULL,   "e",	"emacs",	"emacs",        0,  0 },
  { NULL,   "f",	"LaTeX",	"latex",        1,  0 },
  { NULL,   "f",	"Plain",	"tex",	        1,  0 },
  { NULL,   "f",	"SLiTeX",	"slitex",       1,  0 },
  { NULL,   "v",	"xdvi",		"xdvi",         0,  0 },
  { NULL,   "v",	"ghostview",	"ghostrun",     0,  0 },
  { NULL,   "d",	"dvips",	"dvips",        1,  0 },
  { NULL,   "d",	"dvilj",	"dvilj",        1,  0 },
  { NULL,   "l",	"browse",    	NULL,           0,  0 }  };

_cmd	      *format_cmd,
	      *view_cmd,
	      *edit_cmd,
	      *print_cmd,
	      *log_cmd,
              *bibtex_cmd = &cmd[0],    /* Achtung auf Index */
              *xfig_cmd = &cmd[1];      /* Achtung auf Index */

XtAppContext  app;
extern void   reset_button( Widget w );

/* isnum gibt True zurck, wenn alle character Ziffern sind, sonst False */
/* Leerstring gilt als 0, also True */

Boolean isnum( char *s)
{
  int i;
  for (i=0; i<strlen(s); i++) if (!isdigit(s[i])) return 0;
  return 1;
}

/* a_eq_b vergleicht Strings unabhngig von Gro-/Kleinschreibung;
 * vergleicht nur bis zur Lnge von a, gibt also auch True, wenn
 * der Substring der Lnge a = b ist */

Boolean a_eq_b( char *a, char *b)
{
  int i;
  if (a == NULL && b == NULL) return 1;
/*  if (a == NULL || b == NULL) return 0;*/   /* nur einer NULL => falsch */

  for (i=0; a[i]!=0; i++) if (tolower(a[i])!=tolower(b[i])) return 0;
  return 1;
}

/*-----------------------------------------------------------------*
 * strip gibt einen pointer auf das erste nicht c in str zurck,
 * das erste c in str wird auf \0 gesetzt. Enthlt str nur c's
 * oder \0, wird NULL zurckgegeben. 
 *-----------------------------------------------------------------*/

char *strip( char * str, char c )
{
  int i = 0;
  if (str != NULL)
    { while (str[0] == c) str++;
      while (str[i] != c && str[i] != 0) i++;
      str[i] = 0; }
  if (str[0]) return str;
  else return NULL;
}

/*------------------------------------------------------------------------*
 * concat verkettet beliebig viele Strings (char *)
 * der letzte mu NULL sein
 * Speicher wird automatisch reserviert, mu mit XtFree freigegeben werden
 * first == NULL ist erlaubt, sonst knnten wir schlecht loops benutzen
 *------------------------------------------------------------------------*/

char * concat ( char *first, ... )
{
  va_list lp;
  char *s, *out;

  out = XtMalloc(strlen(first)+1);
  strcpy(out,first);

  va_start(lp,first);
  while((s = va_arg(lp,char *)) != NULL)
    { out = XtRealloc(out,strlen(out)+strlen(s)+1); strcat(out,s); }
  va_end(lp);
  return out;
}

/* Label eines Optionmens in der Struktur cmd suchen und den richtigen	*/
/* Pointer zurckgeben							*/

_cmd *translate_string( char *s )
{
  int i, count;
  count = XtNumber(cmd);			  	          /* Anzahl */
  for (i=0;i<count;i++) if (!strcmp(s,cmd[i].in)) return &cmd[i];
  printf("\nsystem error in translate_string\n"); exit(1);
}

/* Kommando und Parameter konkatenieren und zur Ausfhrung an run bergeben */
/* gibt den string fr die msgbox zurck (mit Kommandoname)                 */

char *run_cmd( _cmd *cmd, char *parm, char *msg )
{
  char *arg[10];
  int i = 0;
  if (cmd->xterm)	/* dann lassen wir's unter aixterm laufen */
    {
      arg[i++] = app_data.xterm; arg[i++] = "-T"; arg[i++] = cmd->in;
      arg[i++] = "-e"; arg[i++] = "run_texshell_cmd";
      if (msg != NULL) { arg[i++] = "-m"; arg[i++] = msg; }
    }

  /* damit's klappt, darf parm kein Blank enthalten
   * (stimmt scheint's gar nicht)
   */
  arg[i++] = cmd->out; arg[i++] = parm; arg[i] = NULL;

  switch (cmd->pid = fork())
    {
    case 0:		  	        /* dies luft im Sohn */
      execvp(arg[0], arg);	        /* wenn alles glatt ging,
				         * ist hier Ende im Sohn */
      perror(arg[0]);                   /* irgendein Fehler */
      exit(1);                          /* der Sohn MUSS beendet werden,
					   sonst wollen 2 den Server managen */
    case -1:
      printf("fork() failed\n");
    }
			                  /* hier luft der Vater weiter */
  if (cmd->w == NULL) printf("Systemfehler in run_cmd\n");
  else if (cmd->pid > 0) XtVaSetValues(cmd->w, XmNsensitive, False, NULL);

  return concat(START_MSG_1[msg_lg],cmd->in,START_MSG_2[msg_lg],NULL);
}

/* macht die menubar Knpfe wieder scharf, wenn das von	*/ 
/* run_cmd gestartete Kind zuende ist (durch SIGCHLD)	*/

void reset()
{
  int pid, i;
#ifdef SYSV
  int status;
#elif AIXV3
  int status;
#else
  union wait status;
#endif
/* signal mu neu gesetzt werden, weil es beim Auftreten auf default	*/
/* zurckgesetzt wurde; geht aber erst nach wait (?), weil sonst alles	*/
/* abstrzt (vielleicht eine Endlosschleife von signals? wei nicht)	*/ 

  if ((pid = wait(&status)) == -1)	         /* irgendein Fehler	*/
    { signal(SIGCHLD,reset);	                 /* (fork ging nicht?), */
      return; }			                 /* ignorieren          */

  for (i = 0; i < XtNumber(cmd); i++)
    if (cmd[i].pid == pid)		         /* fertig, rcksetzen	*/
      { XtAppAddTimeOut(app, 0, reset_button, cmd[i].w);
	signal(SIGCHLD,reset);
	return; }

  /* sollte hier nicht hinkommen: */
  printf("Wieso war Process Nr. %d unser Kind?\n",pid);
}

/*  -- See if file is a pathname from tilda */

char * get_path(char *file)
{
  char  user[ BUFLEN ];
  char* from = &file[1];		/* skip the tilda */
  char* to   = &user[0];

  /* Copy user's name into user[].  I.e., all characters between
   * the `~' and the first `/`.  This may mean no characters (a NUL)
   * which means we will need to look up the user's name.  We will
   * not overflow the user[] array because we already checked the 
   * lenght of name at the beginning.  `from' will point into `name'
   * after the `~user', right at the `/`
   */
  while (*from && *from != '/')
    *to++ = *from++;
  *to = '\0';

  return concat( WcHomeDirectory(user), from, NULL );
}


/*-----------------------------------------------*
 * read config file for dvips or dvilj/2/p;
 * records are:
 *   1.   * driver
 *       (i.e. a comment with exactly one blank followed by the driver's name,
 *        read only for Laserjet)
 *   2-n. driver options, last may be followed by a pipe
 * comments and blank lines maybe freely mixed into the file;
 *
 * data is written to a new printer_list entry;
 * returns address of the new printer
 *-----------------------------------------------*/

Printer *create_printer(FILE *file, char *name)
{
  char    in[BUFSIZ];
  Printer *new_printer = XtNew(Printer);
  last_printer_addr->next = new_printer;       /* Adresse des neuen
					        * im letzten eintragen */
  new_printer->name = name;                    /* Namen merken */
  new_printer->next = NULL;                    /* noch keine folgende  */
  new_printer->driver = "dvips";               /* damit's auch bei PS stimmt */
  new_printer->pipe = NULL;                    /* default */
  last_printer_addr = new_printer;             /* ist jetzt der letzte */

  fgets(in,BUFSIZ,file);                       /* 1. Zeile ist bei LJ driver */
  if (!feof(file) && !strncmp(in,"* dvilj",7))
    { char *d = XtMalloc(strlen(in)-1);
      new_printer->driver = strcpy(d, strtok(in+2,"\n"));
      while (!feof(file) && fgets(in,BUFSIZ,file))           /* Leseschleife */
	{
	  char *line,
	       *opt;
	  if (in[0] == '*') continue;   /* Komentarzeilen werden ignoriert */
	  line = strtok(in,"\n");
	  if (line == NULL)
	    opt = NULL;
	  else
	    opt  = strtok(line,"|");    /* was nach einem Pipe Zeichen steht,
					 * brauchen wir extra zum hintendran
					 * hngen. Darf nur einmal im File
					 * auftreten */
	  if (opt == NULL)
	    new_printer->parms = concat(new_printer->parms," ",line,NULL);
	  else
	    { new_printer->parms = concat(new_printer->parms," ",opt,NULL);
	      new_printer->pipe = concat(strtok(NULL,"|"), NULL); }
	}                 /* mit concat, weil wir Speicher fr pipe brauchen */
    }
  fclose(file);
  return new_printer;
}


/*---------------------------------------------------*
 * aktuelle printer Daten schreiben.
 * liest bei Bedarf Config file fr dvips oder dvilj;
 *---------------------------------------------------*/

Printer *getPrinterData(char *prt, Boolean ok)
{
  Printer *addr = &printer_list[0];
  ok = True;                        /* nehmen wir erst mal an */
  do
    if (!strcmp(prt,addr->name)) return addr;   /* das ist er */
  while ((addr = addr->next) != NULL);

  /* keiner da, config File suchen */
  {
    FILE *file;
    char *fn,
         *dir,
         *default_path = "~:~/dvips.config:~/dvilj.config",
         *path = getenv("TEXCONFIG");

    if (WcNull(path)) path = default_path; 

    dir = strtok(path,":");
    do
      { if (dir[0] ==  '~') dir = get_path(dir);           /* Tilde auflsen */
	fn = concat(dir,"/config.",prt,NULL);
	file = fopen(fn,"r");
	free(fn);
	if (file) return create_printer(file, prt);      /* hier ham wir ihn */
	dir = strtok(NULL,":"); }                      /* nchstes directory */
    while (dir != NULL);
  }
  /* nix brauchbares gefunden, Meldung und default zurckgeben */
  {
    char *msg = concat(INV_CONF_FILE1[msg_lg],prt,INV_CONF_FILE2[msg_lg],NULL);
    start_msgDialog(msg, NULL);
    XtFree(msg);
    ok = False;
    return &printer_list[0];
  }
}
